from .common import ClassifierFCN, ClassifierFCN3D, ClassifierMLP
from .heads_1D import Head1D
from .heads_3D import Head3D
from .head_dense_segmentation import HeadDenseSegmentation
from .head_global_pooling_classifier import HeadGlobalPoolingClassifier
